import { parser } from "./parser.js"
import { foldNodeProp, indentNodeProp, delimitedIndent, LRLanguage } from '@codemirror/language';
import { NodeProp } from '@lezer/common'
import { styleTags, tags, Tag } from "@lezer/highlight"

export const edgeSpecifier = Tag.define(tags.modifier);
export const portSpecifier = Tag.define(tags.modifier);
export const netTypeName = Tag.define(tags.typeName);
export const taskName = tags.function(tags.name);
export const primitiveType = Tag.define(tags.typeName);
export const moduleName = Tag.define(tags.className);
export const attribute = Tag.define(tags.annotation);
export const systemIdentifier = tags.special(tags.variableName);

const tagAssociations = {
    /* Keywords */
    Keyword: tags.keyword,
    'EdgeSymbol! LevelSymbol/Identifier!': edgeSpecifier,
    'localparam parameter': tags.definitionKeyword,
    'module macromodule endmodule': tags.moduleKeyword,
    'function endfunction task endtask': tags.definitionKeyword,
    'primitive endprimitive': tags.definitionKeyword,
    'begin end': tags.keyword,
    /* Sensitivities */
    '@': tags.special(tags.operator),
    'edge posedge negedge': edgeSpecifier,
    /* Control keywords */
    'always initial': tags.controlKeyword,
    'for while': tags.controlKeyword,
    'if else': tags.controlKeyword,
    'case casex casez endcase': tags.controlKeyword,
    /* Primitives and stuff */
    'cmos rcmos': tags.standard(primitiveType),
    'bufif0 bufif1 notif0 notif1': tags.standard(primitiveType),
    'nmos pmos rnmos rpmos': tags.standard(primitiveType),
    'and nand or nor xor xnor': tags.standard(primitiveType),
    'supply0 supply1 tri triand trior tri0 tri1 uwire wire wand wor reg': netTypeName,
    'integer real genvar': tags.typeName,
    /* Modifiers */
    'input output inout': portSpecifier,
    'signed unsigned': tags.modifier,
    /* Names */
    Identifier: tags.variableName,
    'AttrName!': attribute,
    'SystemIdentifier': systemIdentifier,
    'VariableIdentifier!': tags.variableName,
    'DirectiveName!': tags.processingInstruction,
    'UDPIdentifier': primitiveType,
    'UDPDeclaration/UDPIdentifier!': tags.definition(primitiveType),
    'ModuleIdentifier!': moduleName,
    'ModuleDeclaration/ModuleIdentifier!': tags.definition(moduleName),
    'ModuleInstanceIdentifier! UDPInstanceIdentifier!': tags.definition(tags.name),
    'SystemTaskIdentifier! SystemFunctionIdentifier!': tags.function(systemIdentifier),
    'ParamAssignment/ParameterIdentifier!': tags.definition(tags.constant(tags.name)),
    'VariableIdentifier! OutputPortIdentifier!': tags.variableName,
    'r R f F p P n N b B': tags.variableName,
    /* Constant literals */
    UnsignedNumber: tags.number,
    RealNumber: tags.float,
    DecimalValue: tags.integer,
    BinaryValue: tags.integer,
    OctalValue: tags.integer,
    HexValue: tags.integer,
    TimescaleValue: tags.number,
    'x X z Z': tags.number,
    String: tags.string,
    StringEscape: tags.escape,
    'CaseItem/default CaseGenerateItem/default': tags.special(tags.labelName),
    /* Comments */
    'BlockComment BlockComment/"*/"': tags.blockComment,
    LineComment: tags.lineComment,
    'DocComment DocComment/"*/"': tags.docComment,
    /* Brackets */
    '( )': tags.paren,
    '[ ]': tags.angleBracket,
    '(* *)': tags.special(tags.paren),
    '{ }': tags.brace,
};

/**
 * 
 * @template T
 * @template U
 * @param {(key: keyof T, value: T[key], object: T) => U} func 
 * @param {T} obj 
 * @returns {{ [key: string]: U }}
 */
const transform = (func, obj) => {
    /** @type {{ [key: string]: U }} */
    const dst = {};
    for (const key in obj)
        dst[key] = func(key, obj[key], obj);
    return dst;
}

/**
 * @type {{ [open: string]: string }}
 */
export const foldingPairs = {
    'module': 'endmodule',
    'macromodule': 'endmodule',
    'casex': 'endcase',
    'casez': 'endcase',
    'case': 'endcase',
    'begin': 'end',
    'fork': 'join',
    'config': 'endconfig',
    'table': 'endtable',
    'generate': 'endgenerate',
    'function': 'endfunction',
    'primitive': 'endprimitive',
    'task': 'endtask',
    '(*': '*)',
    '/*': '*/',
    '/**': '*/',
    '(': ')',
    '[': ']',
    '{': '}'
};

export const folding = foldNodeProp.add({
    [
        'ModuleDeclaration'
        + ' ConfigDeclaration'
        + ' FunctionDeclaration'
        + ' TaskDeclaration'
        + ' UDPDeclaration'
    ]: context => {
        const sc = context.getChild(';');
        const lc = context.lastChild;
        if (!sc || !lc)
            return null;

        return { from: sc.to, to: lc.from - 1};
    },
    [
        'GenerateRegion'
        + ' UDPBody'
        + ' SpecifyBlock'
    ]: context => {
        const fc = context.firstChild;
        const lc = context.lastChild;
        if (!fc || !lc)
            return null;

        return { from: fc.to, to: lc.from };
    },
    'CaseGenerateConstruct CaseStatement': context => {
        const fc = context.getChild(')');
        const lc = context.lastChild;
        if (!fc || !lc)
            return null;

        return { from: fc.to, to: lc.from };
    },
    'GenerateBlock': context => {
        const fc = context.getChild('GenerateBlockIdentifier')
                ?? context.firstChild;
        const lc = context.lastChild;
        if (!fc || !lc)
            return null;

        return { from: fc.to, to: lc.from };
    },
    'ParBlock SeqBlock': context => {
        const fc = context.getChild('BlockIdentifier') ?? context.firstChild;
        const lc = context.lastChild;
        if (!fc || !lc)
            return null;

        return { from: fc.to, to: lc.from };
    }
});
export const closedBy = NodeProp.closedBy.add(transform((_, close) => [close], foldingPairs))
export const styling = styleTags(tagAssociations);

const augmentedParser = parser.configure({
    props: [ 
        styling, 
        closedBy,
        folding,
        indentNodeProp.add({
            'PortDeclarationList PortList ModuleParameterPortList'
                : delimitedIndent({ closing: ')', align: true }),
            'ModuleItem' : ctx => ctx.textAfter.match(/^(case|begin|fork)/)
                                ? ctx.baseIndent 
                                : ctx.baseIndent + ctx.unit,
            'SeqBlock' : delimitedIndent({ closing: 'end', align: true }),
            'GenerateRegion' : delimitedIndent({ closing: 'endgenerate', align: true }),
            'ParBlock' : delimitedIndent({ closing: 'join', align: true }),
            'ModuleItemList' : delimitedIndent({ closing:'endmodule'})
        }),
    ]
})

export const verilogLanguage = LRLanguage.define({
    parser: augmentedParser,
    languageData: {
        'closeBrackets': {
            brackets: Object.keys(foldingPairs)
        }
    }
});

export default verilogLanguage;